/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"

// global defines :  RTAVR_INCLUDE_IOR, RTAVR_INCLDUE_RAM, RTAVR_INCLUDE_ROM

`define RTAVR_INCLUDE_GPR
//`define RTAVR_INCLUDE_IOR
//`define RTAVR_INCLUDE_RAM
//`define RTAVR_INCLUDE_ROM
`define RTAVR_INCLUDE_S0
`define RTAVR_INCLUDE_S1

//
// Module Name: rtavr
//   ver 002
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                                        all-in-one  
// Number of Slice Flip Flops:              247
// Number of 4 input LUTs:                  722
// Number of occupied Slices:               472
// Total Number of 4 input LUTs:            762
// Number of bonded IOBs :                   26
//   IOB Flip Flops :                        24
// Number of BUFGMUXs :                       2
// Number used for Dual Port RAMs            16
// Number of RAMB16BWEs :                     2
// Clock to Setup 
//on destination clock CLK2X              13.030

module rtavr (
		input RESET
`ifdef RTAVR_INCLUDE_S0
`else
		, input S0_WB_INST
		, output S0_WB_EN
`endif
`ifdef RTAVR_INCLUDE_S1
`else
		, input CMD_LOAD, CMD_LOAD_WR, CMD_STORE
		, input CMD_OP3, CMD_OP2, CMD_OP1
		, input [2:0] OP3_SEL
		, input OP3_WC, CMD_OP3_WR
		, input [2:0] OP2_SEL
		, input [2:0] OP1_SEL
		, input [7:0] IMM_DATA
		, output [15:0] INST

		, input CMD_SBIX;
		, input CMD_SBRX;
		, input [2:0] SBIX_BIT;
		, output SBIX_BIT_OUT;

		, input EA_ABS
		, input [7:0] ABS_ADDR
		, input [11:0] PC
//		, output FORWARD_COND
`endif

`ifdef RTAVR_INCLUDE_GPR
		, input CLK2X
  `ifdef USE_DMY_CLOCK
  `else
		, input CLK
  `endif
`else
		, input CLK
		, output GPR_WE
		, output[7:0] GPR_DI
		, input[7:0] GPR_DOAL
		, input[7:0] GPR_DOAH
		, input[7:0] GPR_DOBL
		, input[7:0] GPR_DOBH
		, output[12:0] GPR_JA
		, output GPR_PREDEC, GPR_POSTINC
		, output[3:0] GPR_ADDRAL
		, output[3:0] GPR_ADDRAH
		, output[3:0] GPR_ADDRBL
		, output[3:0] GPR_ADDRBH
`endif
`ifdef RTAVR_INCLUDE_IOR
		// external I/O
  `ifdef IOR_HAVE_PORTA
		, inout [7:0] PORTA
  `endif
  `ifdef IOR_HAVE_PORTB
		, inout [7:0] PORTB
  `endif
  `ifdef IOR_HAVE_PORTC
		, inout [7:0] PORTC
  `endif
`else
		, output IOR_WEB, IOR_OE
		, output [5:0] IOR_ADDRB
		, output [7:0] IOR_DIB
		, input [7:0] IOR_DOB

		// SREG(FLAGS) & SP
		, input [7:0]; FLAGS_IN
		, output [7:0]; FLAGS_OUT
		, input [15:0]; SP_IN
		, output [15:0]; SP_OUT
`endif
`ifdef RTAVR_INCLUDE_RAM
`else
		, output RAM_WEB, RAM_OE
		, output [10:0] RAM_ADDRB
		, output [7:0] RAM_DIB
		, input [7:0] RAM_DOB
`endif
`ifdef RTAVR_INCLUDE_ROM
`else
		, output ROM_WEB,ROM_OE
		, output [12:0] ROM_ADDRB
		, output [7:0] ROM_DIB
		, input[7:0] ROM_DOB
`endif
		);


`ifdef RTAVR_INCLUDE_GPR
  `ifdef USE_DMY_CLOCK
  wire CLK;
  `endif

  wire GPR_WE;
  wire [7:0] GPR_DI;
  wire GPR_PREDEC,GPR_POSTINC;
  wire [3:0] GPR_ADDRAL;
  wire [3:0] GPR_ADDRAH;
  wire [3:0] GPR_ADDRBL;
  wire [3:0] GPR_ADDRBH;
  wire [7:0] GPR_DOAL;
  wire [7:0] GPR_DOAH;
  wire [7:0] GPR_DOBL;
  wire [7:0] GPR_DOBH;
  wire [11:0] GPR_JA;
  wire WB_VALID;
  rtavr_gpr_16 GPR(.CLK2X(CLK2X), .RESET(RESET)
`ifdef USE_DMY_CLOCK
		,  .CLK_OUT(CLK)
`else
		,  .CLK(CLK)
`endif
		, .WE(GPR_WE), .DI(GPR_DI)
                , .PREDEC(GPR_PREDEC), .POSTINC(GPR_POSTINC)
		, .ADDRAL(GPR_ADDRAL[3:0]), .ADDRAH(GPR_ADDRAH[3:0])
		, .ADDRBL(GPR_ADDRBL[3:0]), .ADDRBH(GPR_ADDRBH[3:0])
		, .DOAL(GPR_DOAL[7:0]), .DOAH(GPR_DOAH[7:0])
		, .DOBL(GPR_DOBL[7:0]), .DOBH(GPR_DOBH[7:0])
		, .JA(GPR_JA[11:0])
		, .WB_VALID(WB_VALID)
		); 
`endif
`ifdef RTAVR_INCLUDE_IOR
  wire IOR_WEB, IOR_OE;
  wire [5:0] IOR_ADDRB;
  wire [7:0] IOR_DIB;
  wire [7:0] IOR_DOB;

  wire [7:0] FLAGS_IN;
  wire [7:0] FLAGS_OUT;
  wire [15:0] SP_IN;
  wire [15:0] SP_OUT;
  
  wire IOR_RST_INT0 = 0;
  wire IOR_RST_TIM0_OVF = 0;
  wire IOR_RST_TIM0_COMPA = 0;
  wire IOR_RST_TIM0_COMPB = 0;

  rtavr_ior IOR (.CLK(CLK), .RESET(RESET), .WR(IOR_WEB), .RD(IOR_OE)
		, .ADDR(IOR_ADDRB[5:0]), .DI(IOR_DIB[7:0]), .DO(IOR_DOB[7:0])
  `ifdef IOR_HAVE_PORTA
		, .PORTA(PORTA[7:0])
  `endif
  `ifdef IOR_HAVE_PORTB
		, .PORTB(PORTB[7:0])
  `endif
  `ifdef IOR_HAVE_PORTC
		, .PORTC(PORTC[7:0])
  `endif
		// SREG(FLAGS) & SP
		, .FLAGS_IN(FLAGS_OUT[7:0]) , .FLAGS_OUT(FLAGS_IN[7:0])
		, .SP_IN(SP_OUT[15:0]) , .SP_OUT(SP_IN[15:0])
		// INTERRUPTS 
		, .RST_INT0(IOR_RST_INT0)
		, .INT0(IOR_INT0)
		, .RST_TIM0_OVF(IOR_RST_TIM0_OVF)
		, .TIM0_OVF(IOR_TIM0_OVF)
		, .RST_TIM0_COMPA(IOR_RST_TIM0_COMPA)
		, .TIM0_COMPA(IOR_TIM0_COMPA)
		, .RST_TIM0_COMPB(IOR_RST_TIM0_COMPB)
		, .TIM0_COMPB(IOR_TIM0_COMPB)
		);
`endif
`ifdef RTAVR_INCLUDE_RAM
  wire RAM_WEB, RAM_OE;
  wire [10:0] RAM_ADDRB;
  wire [7:0] RAM_DIB;
  wire [7:0] RAM_DOB;

  wire RAM_WEA = 0;
  wire [10:0] RAM_ADDRA = 0;
  wire [7:0] RAM_DIA = 0;
 
  rtavr_sram_2KB RAM(.CLK(CLK), .RESET(RESET), .WEB(RAM_WEB), .OE(RAM_OE)
		, .ADDRB(RAM_ADDRB[10:0])
		, .DIB(RAM_DIB[7:0]), .DOB(RAM_DOB[7:0])
		, .WEA(RAM_WEA), .DIA(RAM_DIA[7:0]), .ADDRA(RAM_ADDRA[10:0])
		);
`endif
`ifdef RTAVR_INCLUDE_ROM
  wire ROM_WEB, ROM_OE;
  wire [12:0] ROM_ADDRB;
  wire [7:0] ROM_DIB;
  wire [7:0] ROM_DOB;
  
  wire [11:0] ROM_ADDRA;
  wire [15:0] ROM_DOA;
  rtavr_rom #(.SIZE(`ROM_SIZE))
	ROM (.CLK(CLK), .RESET(RESET), .WEB(ROM_WEB), .OE(ROM_OE)
		, .ADDRB(ROM_ADDRB[12:0])
		, .DIB(ROM_DIB[7:0]), .DOB(ROM_DOB[7:0])
		, .ADDRA(ROM_ADDRA[11:0]), .DOA(ROM_DOA[15:0])
		);
`endif
`ifdef RTAVR_INCLUDE_S0
  wire [15:0] INST;
  wire S0_WB_INST;
  wire S0_WB_EN;
  rtavr_s0_fetch  S0 (.CLK(CLK), .RESET(RESET), .PM_OUT(ROM_DOA[15:0])
                , .WB_EN(S0_WB_EN)
                , .WB_INST(S0_WB_INST)
		, .INST(INST[15:0])
                , .GPR_PREDEC(GPR_PREDEC), .GPR_POSTINC(GPR_POSTINC)
		, .GPR_ADDRAL(GPR_ADDRAL[3:0]), .GPR_ADDRAH(GPR_ADDRAH[3:0])
		, .GPR_ADDRBL(GPR_ADDRBL[3:0]), .GPR_ADDRBH(GPR_ADDRBH[3:0])
		);
`endif
`ifdef RTAVR_INCLUDE_S1
  wire [2:0] OP3_SEL;
  wire [2:0] OP2_SEL;
  wire [2:0] OP1_SEL;
  wire [7:0] IMM_DATA;
  wire [7:0] ABS_ADDR;
  wire [11:0] PC;
  wire [1:0] PUSH_SEL;

  wire [2:0] SBIX_BIT;

  rtavr_s1_decode S1 (.CLK(CLK), .RESET(RESET)
                , .S0_WB_INST(S0_WB_INST)
                , .S0_WB_EN(S0_WB_EN)
		, .WB_VALID(WB_VALID)
                , .INST(INST[15:0])
                , .FLAGS_IN(FLAGS_OUT[7:0])
                , .JA(GPR_JA[11:0])

                , .CMD_LOAD(CMD_LOAD)
                , .CMD_LOAD_WR(CMD_LOAD_WR)
		, .CMD_STORE(CMD_STORE)
		, .CMD_OP3(CMD_OP3), .CMD_OP2(CMD_OP2), .CMD_OP1(CMD_OP1)
		, .CMD_OP3_WR(CMD_OP3_WR)
		, .OP3_WC(OP3_WC)
		, .OP3_SEL(OP3_SEL[2:0])
		, .OP2_SEL(OP2_SEL[2:0])
		, .OP1_SEL(OP1_SEL[2:0])
		, .IMM_DATA(IMM_DATA[7:0])

		, .CMD_SBIX(CMD_SBIX)
		, .CMD_SBRX(CMD_SBRX)
		, .SBIX_BIT(SBIX_BIT[2:0])
		, .SBIX_BIT_IN(SBIX_BIT_OUT)

		, .EA_ABS(EA_ABS), .ABS_ADDR(ABS_ADDR[7:0])
		, .EA_PUSH(EA_PUSH), .EA_POP(EA_POP)
		, .PUSH_SEL(PUSH_SEL[1:0])
		, .PC(PC[11:0])
		);
`endif
  assign SP_OUT[15:12] = SP_IN[15:12];
  assign SP_OUT[11:0] = EA_PUSH ? SP_IN[11:0] + 1
                        : EA_POP  ? SP_IN[11:0] - 1
		        : SP_IN[11:0];

  wire [15:0] EA = EA_PUSH ? { 4'b0000, SP_OUT }
                 : EA_POP  ? { 4'b0000, SP_IN }
		 : EA_ABS  ? { 8'b00000000, ABS_ADDR }
		 : { GPR_DOAH , GPR_DOAL };
  
  assign ROM_ADDRA = PC;
  
  // 0x0000 - 0x003f (64B)
  wire oe_ior = ((~EA_ABS & (EA[15:8] == 0)) | EA_ABS ) & (EA[7:6] == 0);

  // 0x0040 - 0x1000 (4KB cyclic)
  wire oe_ram =  (~EA_ABS & (EA[15:12] == 0)  & ~(EA[11:6] == 0))
                  | (EA_ABS & ~(EA[7:6] == 0)) ;

  // 0x4000 - 0x5fff (8KB)
  wire oe_rom = ~EA_ABS & (EA[15:13] == 2);
    
  wire wr_ior = CMD_STORE & oe_ior;
  wire wr_ram = CMD_STORE & oe_ram;
  wire wr_rom = CMD_STORE & oe_rom;

  wire rd_ior = CMD_LOAD & oe_ior;
  wire rd_ram = CMD_LOAD & oe_ram;
  wire rd_rom = CMD_LOAD & oe_rom;
  
  wire [7:0] Rr_in = GPR_DOBL;
  wire [7:0] Rd_in = GPR_DOBH;

  wire C_out;
  wire Z_out;
  wire N_out;
  wire V_out;
  wire S_out;
  wire H_out;

  wire C_in = FLAGS_IN[0];
  wire Z_in = FLAGS_IN[1];
  wire N_in = FLAGS_IN[2];
  wire V_in = FLAGS_IN[3];
  wire S_in = FLAGS_IN[4];
  wire H_in = FLAGS_IN[5];
  
  assign FLAGS_OUT[0] = C_out;
  assign FLAGS_OUT[1] = Z_out;
  assign FLAGS_OUT[2] = N_out;
  assign FLAGS_OUT[3] = V_out;
  assign FLAGS_OUT[4] = S_out;
  assign FLAGS_OUT[5] = H_out;
  assign FLAGS_OUT[7:6] = FLAGS_IN[7:6];

// OP3          --*-:**------------                               HSVNZC
// CPSE		000B:00rd:dddd:rrrr	# B=1: CPSE           #   ------
// CP/CPC       000B:01rd:dddd:rrrr	# B=1: CP  B=0: CPC   #   XXXXXX
// SUB/SBC	000B:10rd:dddd:rrrr	# B=1: sub B=0: sbc   #   XXXXXX
// ADD/ADD	000C:11rd:dddd:rrrr	# C=0: add C=1: adc   #   XXXXXX
// AND		0010:00rd:dddd:rrrr	# logical AND         #   -X0XX-
// EOR		0010:01rd:dddd:rrrr	# exclusive OR        #   -X0XX-
// OR		0010:10rd:dddd:rrrr	# inclusive OR        #   -X0XX-
// MOV		0010:11rd:dddd:rrrr	# copy register       #   ------

  wire [8:0] OP3_OUT = (OP3_SEL == 7) ? {0, Rr_in }
		: (OP3_SEL == 6) ? {0, (Rd_in | Rr_in) }
		: (OP3_SEL == 5) ? {0, (Rd_in ^ Rr_in) }
		: (OP3_SEL == 4) ? {0, (Rd_in & Rr_in) }
		: (OP3_SEL == 3) ? ({0, Rd_in}  + {0, Rr_in}
				   + {8'b0, (OP3_WC)? C_in : 0 })
		: ({0,Rd_in} - {0, Rr_in} - {8'b0, (OP3_WC) ? 0: C_in }) ;
  wire op3_v   =  ~OP3_SEL[2]
               ?   OP3_OUT[7] & ~Rd_in[7] & ~Rr_in[7]
                | ~OP3_OUT[7] &  Rd_in[7] &  Rr_in[7]
	       : (OP3_SEL == 7) ? V_in
	       : 0 ;
  wire op3_h   =  ~OP3_SEL[2]
               ?  OP3_OUT[3] & ~Rd_in[3] 
                | OP3_OUT[3] & ~Rr_in[3]
                | Rd_in[3] & Rr_in[3]
	       : H_in;

// # OP2 INST
//              -***:--------------                           #   HSVNZC
// LDI		1110:KKKK:dddd:KKKK	#                     #   ------
// CPI		0011:KKKK:dddd:KKKK	#                     #   XXXXXX 
// SBCI		0100:KKKK:dddd:KKKK	#                     #   XXXXXX 
// SUBI		0101:KKKK:dddd:KKKK	#                     #   XXXXXX
// ORI		0110:KKKK:dddd:KKKK	#                     #   -X0XX- 
// ANDI		0111:KKKK:dddd:KKKK	#                     #   -X0XX-  
 

  wire [8:0] OP2_OUT = (OP2_SEL == 7) ? {0, (Rd_in & IMM_DATA ) }
		: (OP2_SEL == 6) ? {0, (Rd_in | IMM_DATA) } 
		: (OP2_SEL == 5) ? ({0,Rd_in} - {0, IMM_DATA})
		: (OP2_SEL == 4) ? ({0,Rd_in} - {0, IMM_DATA} - {8'b0, C_in })	
		: (OP2_SEL == 3) ? 0 -  { 1'b0 , IMM_DATA }
		: { C_in , IMM_DATA };
  wire op2_v   = (OP2_SEL == 3) | (OP2_SEL == 4) | (OP2_SEL == 5)
               ?   OP2_OUT[7] & ~Rd_in[7] & ~IMM_DATA[7]
                 | ~OP2_OUT[7] &  Rd_in[7] &  IMM_DATA[7]
	       : (OP2_SEL == 6) | (OP2_SEL == 7) ? 0
	       : V_in;
  wire op2_h   = (OP2_SEL == 3) | (OP2_SEL == 4) | (OP2_SEL == 5) 
               ?   OP2_OUT[3] & ~Rd_in[3] 
                 | OP2_OUT[3] & ~IMM_DATA[3]
                 | Rd_in[3] & IMM_DATA[3]
	       : H_in;

// OP1          --------------:0***                           #   HSVNZC
// COM		1001:010d:dddd:0000	#                     #   -X0XX1
// NEG		1001:010d:dddd:0001	#                     #   XXXXXX 
// SWAP		1001:010d:dddd:0010	#                     #   ------
// INC		1001:010d:dddd:0011	#                     #   -XXXX-
// DEC		1001:010d:dddd:1010	#  (1010 -> 0100 )    #   -XXXX-
// ASR		1001:010d:dddd:0101	#                     #   -XXXXX
// LSR		1001:010d:dddd:0110	#                     #   -XX0XX
// ROR		1001:010d:dddd:0111	#                     #   -XXXXX

  wire [8:0] OP1_OUT = (OP1_SEL == 7) ? { Rd_in , C_in }
		: (OP1_SEL == 6) ? { Rd_in[0] , 1'b0, Rd_in[7:1] }
		: (OP1_SEL == 5) ? { Rd_in[0] , Rd_in[7] , Rd_in[7:1] }
		: (OP1_SEL == 4) ? { C_in , Rd_in - 1 }
		: (OP1_SEL == 3) ? { C_in , Rd_in + 1 }
		: (OP1_SEL == 2) ? { 1'b0 , Rd_in[3:0] , Rd_in[7:4] }
		: (OP1_SEL == 1) ? 0 -  { 1'b0 , Rd_in }
		: (OP1_SEL == 0) ? { 1'b1 , ~Rd_in } : 0;
 
  wire op1_v   =  (OP1_SEL == 3) ? ( Rd_in == 8'b01111111 )  // INC
                : (OP1_SEL == 1) | (OP1_SEL == 4) ?          // NEG, DEC
                                   ( Rd_in == 8'b10000000 )
		: (OP1_SEL == 2) ? V_in                      // SWAP
                : OP1_OUT[7] ^ Rd_in[0];                // ASR , LSR , ROR

  wire op1_h   = (OP1_SEL == 1) ?  OP1_OUT[3] | ~Rd_in[3]      // NEG
                : H_in ;

  wire v_sbrx_bit =  (SBIX_BIT == 0) ? Rr_in[0]
                   : (SBIX_BIT == 1) ? Rr_in[1] 
		   : (SBIX_BIT == 2) ? Rr_in[2]
		   : (SBIX_BIT == 3) ? Rr_in[3]
		   : (SBIX_BIT == 4) ? Rr_in[4]
		   : (SBIX_BIT == 5) ? Rr_in[5]
		   : (SBIX_BIT == 6) ? Rr_in[6]
		   : Rr_in[7];

  wire v_sbri_bit =  (SBIX_BIT == 0) ? IOR_DOB[0]
                   : (SBIX_BIT == 1) ? IOR_DOB[1] 
		   : (SBIX_BIT == 2) ? IOR_DOB[2]
		   : (SBIX_BIT == 3) ? IOR_DOB[3]
		   : (SBIX_BIT == 4) ? IOR_DOB[4]
		   : (SBIX_BIT == 5) ? IOR_DOB[5]
		   : (SBIX_BIT == 6) ? IOR_DOB[6]
		   : IOR_DOB[7];

  assign di_is_zero = (GPR_DI == 0);
  assign SBIX_BIT_OUT = CMD_SBIX ? v_sbri_bit 
                       : CMD_SBRX ? v_sbrx_bit
		       : di_is_zero ;
//		       : (Rd_in == Rr_in) ;

  assign IOR_OE = rd_ior;
  assign RAM_OE = rd_ram;
  assign ROM_OE = rd_rom;

  assign IOR_WEB = wr_ior;
  assign RAM_WEB = wr_ram;
  assign ROM_WEB = wr_rom;

  assign IOR_ADDRB = EA[5:0];
  assign RAM_ADDRB = EA[10:0];
  assign ROM_ADDRB = EA[11:0];
  
  wire [7:0] load_data = oe_ior ? IOR_DOB : oe_ram ? RAM_DOB 
                       : oe_rom ? ROM_DOB : 0;

  wire [7:0] store_data = (EA_PUSH & (PUSH_SEL == 2'b10)) ? PC[7:0]
                        : (EA_PUSH & (PUSH_SEL == 2'b11)) ? { 4'b0, PC[11:8] }
			: GPR_DOBH;
  
  assign IOR_DIB = store_data;
  assign RAM_DIB = store_data;
  assign ROM_DIB = store_data;
  
  assign GPR_DI =  CMD_LOAD ? load_data : CMD_OP3 ? OP3_OUT[7:0] 
                 : CMD_OP2 ? OP2_OUT[7:0] : CMD_OP1 ? OP1_OUT[7:0] : 0 ;
  assign GPR_WE = CMD_LOAD_WR | CMD_OP3_WR | CMD_OP2 | CMD_OP1 ;
  
 
  assign C_out = CMD_OP3_WR ? OP3_OUT[8] : CMD_OP2 ? OP2_OUT[8] 
                 : CMD_OP1 ? OP1_OUT[8] : C_in ; 

  assign Z_out = (CMD_OP3_WR | CMD_OP2 | CMD_OP1) ? di_is_zero : Z_in;
  assign N_out = (CMD_OP3_WR | CMD_OP2 | CMD_OP1) ? GPR_DI[7] : N_in;
  assign S_out = (CMD_OP3_WR | CMD_OP2 | CMD_OP1) ? N_out ^ V_out : S_in;
  
  assign V_out = CMD_OP3_WR ? op3_v : CMD_OP2 ? op2_v : CMD_OP1 ? op1_v : V_in; 
  assign H_out = CMD_OP3_WR ? op3_h : CMD_OP2 ? op2_h : CMD_OP1 ? op1_h : H_in;

//  always @(posedge CLK)
//  begin
//  end

endmodule
