/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_T1, IOR_HAVE_PORTA , IOR_HAVE_PORTB
//                 , IOR_HAVE_PORTC, IOR_HAVE_SREG_SP

//`define OUTPUT_MUX_T1
//`define IOR_HAVE_PORTA
//`define IOR_HAVE_PORTB
//`define IOR_HAVE_PORTC
//`define IOR_HAVE_SREG_SP

`define TIM0_FUNC_MINIMUM


// module : rtavr_ior
//   sub module : rtavr_ior_port
//   sub module : rtavr_ior_timer0
// version 003


// Module Name: rtavr_ior_port
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                          DO:  NORMAL    TRISTATE(FAKE)   
// Number of Slice Flip Flops:     16         16
// Number of 4 input LUTs:         22         23
// Number of occupied Slices:      19         20
// Total Number of 4 input LUTs:   22         23 
// Number of bonded IOBs :         33         33
//   IOB Flip Flops :              16         16
// Number of BUFGMUXs :             1          1
// Number of RAMB16BWEs :          -           -

module rtavr_ior_port(
		input CLK,
		input RESET,

		input [7:0] DI,
		output [7:0] DO
		, input WR,RD
		, input [5:0] ADDR
		// external I/O
		, inout [7:0] PORT
		);
  parameter BASE_ADDR = 0; // PORTA 0x00
//  parameter BASE_ADDR = 4; // PORTB 0x04
//  parameter BASE_ADDR = 27; // PORTC 0x1b

  reg  [7:0] i_pin;
  reg  [7:0] i_ddr;
  reg  [7:0] i_port;

  wire OE_pin = RD & (ADDR == BASE_ADDR);
  //wire WE_pin = WR & (ADDR == BASE_ADDR);
  wire OE_ddr = RD & (ADDR == (BASE_ADDR + 1)); 
  wire WE_ddr = WR & (ADDR == (BASE_ADDR + 1));
  wire OE_port = RD & (ADDR == (BASE_ADDR + 2)); 
  wire WE_port = WR & (ADDR == (BASE_ADDR + 2));

`ifdef OUTPUT_MUX_T1
  wire [7:0] i_DO = OE_ddr ? i_ddr : OE_pin ? i_pin : OE_port ? i_port : 0;
  bufif1(DO, i_DO, (OE_pin | OE_ddr | OE_port));
`else
  wire [7:0] i_DO = OE_ddr ? i_ddr : OE_pin ? i_pin : OE_port ? i_port : 0;
  assign DO = i_DO;
`endif
  always @(negedge CLK)
  begin
    if (RESET == 0)
      begin
	i_pin <= 0;
      end
    else
      begin
	i_pin <= PORT;
      end
  end
  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	i_ddr <= 0;
	i_port <= 0;
      end
    else
      begin
	//if(WE_pin) i_pin <= DI;
	if(WE_ddr) i_ddr <= DI;
	if(WE_port) i_port <= DI;
      end
  end

  generate 
    genvar i;
    for (i=0; i<8; i=i+1)
       begin : port_out
	   bufif1(PORT[i],i_port[i],i_ddr[i]);
       end
  endgenerate
endmodule

// Module Name: rtavr_ior_timer0
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                          DO:  NORMAL    TRISTATE(FAKE)   
// Number of Slice Flip Flops:     49         49   
// Number of 4 input LUTs:         69         73
// Number of occupied Slices:      56         57
// Total Number of 4 input LUTs:   78         82 
// Number of bonded IOBs :         28
//   IOB Flip Flops :              -
// Number of BUFGMUXs :             1
// Number of RAMB16BWEs :          -

module rtavr_ior_timer0(
		input CLK,
		input RESET,

		input [7:0] DI,
		output [7:0] DO

		, input WR,RD
		, input [5:0] ADDR

		// external I/O
		, output TIM0_OVF
		, output TIM0_COMPA	
		, output TIM0_COMPB
		);
  parameter BASE_ADDR = 21; // OCR0B 0x15
  reg  [9:0] i_prescaler;
  reg  [7:0] i_tcnt0;
  reg  [7:0] i_ocr0b;
  reg  [7:0] i_ocr0a;
  reg  i_tim0_compa;
  reg  i_tim0_compb;
  reg  i_tim0_ovf;
 
  assign TIM0_OVF = i_tim0_ovf;
  assign TIM0_COMPA = i_tim0_compa;
  assign TIM0_COMPB = i_tim0_compb;


  wire OE_ocr0b = RD & (ADDR == BASE_ADDR);
  wire WE_ocr0b = WR & (ADDR == BASE_ADDR);
  wire OE_ocr0a = RD & (ADDR == (BASE_ADDR + 1));
  wire WE_ocr0a = WR & (ADDR == (BASE_ADDR + 1));
  wire OE_tcnt0 = RD & (ADDR == (BASE_ADDR + 2));
  wire WE_tcnt0 = WR & (ADDR == (BASE_ADDR + 2));
  wire OE_tccr0b = RD & (ADDR == (BASE_ADDR + 3));
  wire WE_tccr0b = WR & (ADDR == (BASE_ADDR + 3));
  wire OE_tccr0a = RD & (ADDR == (BASE_ADDR + 4));
  wire WE_tccr0a = WR & (ADDR == (BASE_ADDR + 4));

  reg [2:0] cs0;
  reg [2:0] wgm0;
  reg [1:0] com0a;
  reg [1:0] com0b;

  reg tsm;
`ifdef TIM0_FUNC_MINIMUM
`else
  reg psr;
  reg foc0b;
  reg foc0a;
`endif
  wire  [7:0] i_tccr0b = {com0a, com0b, wgm0[2], cs0};
  wire  [7:0] i_tccr0a = {2'b0, tsm, 3'b0 , wgm0[1:0]};

  reg t0_prev;
  wire t0_nops = (cs0 == 1);
  wire t0_clk =  (cs0 == 2) ? i_prescaler[2] // 1/8
               : (cs0 == 3) ? i_prescaler[5] // 1/64
               : (cs0 == 4) ? i_prescaler[7] // 1/256
               : (cs0 == 5) ? i_prescaler[9] // 1/1024
               : 0;
  wire t0_top_ocr0a = ((wgm0 == 2) | (wgm0 == 5) | (wgm0 == 7));
`ifdef TIM0_FUNC_MINIMUM
`else
  wire ocr0_chg_top = ((wgm0 == 1) | (wgm0 == 5));
  wire ocr0_chg_btm = ((wgm0 == 3) | (wgm0 == 7));
  wire ocr0_chg_imm = (wgm0[0] == 0);
`endif
`ifdef OUTPUT_MUX_T1
  wire [7:0] i_DO = OE_ocr0b ? i_ocr0b : OE_ocr0a ? i_ocr0a : OE_tcnt0 ? i_tcnt0
              : OE_tccr0b ? i_tccr0b : OE_tccr0a ? i_tccr0a : 0 ;
  bufif1(DO,i_DO, (OE_ocr0b | OE_ocr0a | OE_tcnt0 | OE_tccr0b | OE_tccr0a) );
`else
  wire [7:0] i_DO = OE_ocr0b ? i_ocr0b : OE_ocr0a ? i_ocr0a : OE_tcnt0 ? i_tcnt0
              : OE_tccr0b ? i_tccr0b : OE_tccr0a ? i_tccr0a : 0 ;
  assign DO = i_DO;
`endif
  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	i_ocr0b <= 0; 	 
	i_ocr0a <= 0;
	cs0 <= 0;
	wgm0[2] <= 0;
	com0b <= 0;
	com0a <= 0;
	wgm0[1:0] <= 0;
	tsm <= 0;
`ifdef TIM0_FUNC_MINIMUM
`else
	psr <= 0;
	foc0b <= 0;
	foc0a <= 0;
`endif
	i_prescaler <= 0;
	i_tcnt0 <= 0;
	i_tim0_ovf   <= 0;
	i_tim0_compa <= 0;
	i_tim0_compb <= 0;
	t0_prev <= 0;
      end
    else
      begin
	if(WE_ocr0b) i_ocr0b <= DI; 	 
	if(WE_ocr0a) i_ocr0a <= DI;
	if(WE_tccr0b) 
	begin
	  cs0 <= DI[2:0];
	  wgm0[2] <= DI[3];
	  com0b <= DI[5:4];
	  com0a <= DI[7:6];
	end
	if(WE_tccr0a) 
	begin
	  wgm0[1:0] <= DI[1:0];
	  tsm <= DI[5];
`ifdef TIM0_FUNC_MINIMUM
`else
	  psr <= DI[4];
	  foc0b <= DI[6];
	  foc0a <= DI[7];
`endif
	end
`ifdef TIM0_FUNC_MINIMUM
	i_prescaler <= i_prescaler + 1;
	if(WE_tcnt0) i_tcnt0 <= DI;
	else
	begin
	  if (t0_nops | (t0_clk ^ t0_prev))
            begin
		if (t0_top_ocr0a & (i_tcnt0 == i_ocr0a)) i_tcnt0 <= 0; 
		else i_tcnt0 <= i_tcnt0 + 1;
 
		i_tim0_ovf   <= (i_tcnt0 == 255);
		i_tim0_compa <= (i_tcnt0 == i_ocr0a);
		i_tim0_compb <= (i_tcnt0 == i_ocr0b);
	    end
	end
`else
NOT IMPLEMENTED YET
`endif

	t0_prev <= t0_clk;
      end
  end
endmodule

// Module Name: rtavr_ior
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//                              NO_PORT   +PORTA +PORTB +PORTC +SREG/SP
// Number of Slice Flip Flops:     58       78     94   110   134(136)
// Number of 4 input LUTs:         90       120   142   153   215(223)
// Number of occupied Slices:      77       104   122   137   169(175)
// Total Number of 4 input LUTs:   99       129   151   162   224(232)
// Number of bonded IOBs :         31        44    49    57   105(106)
//   IOB Flip Flops :               -        15    31    47    71
// Number of BUFGMUXs :             1         1     1     1     1
// Number of RAMB16BWEs :           -         -     -     -     -
// CLK to Setup                  5.845                        5.951(6.264)
// () -- add INT0 settings / RESET

module rtavr_ior (
		input CLK,
		input RESET,

		input WR,RD, 
		input [7:0] DI,
		input [5:0] ADDR,
		output [7:0] DO

`ifdef IOR_HAVE_SREG_SP
		, input  [7:0] FLAGS_IN
		, input  [15:0] SP_IN
		, output [7:0] FLAGS_OUT
		, output [15:0] SP_OUT
`endif
		// external I/O
`ifdef IOR_HAVE_PORTA
		, inout [7:0] PORTA
`endif
`ifdef IOR_HAVE_PORTB
		, inout [7:0] PORTB
`endif
`ifdef IOR_HAVE_PORTC
		, inout [7:0] PORTC
`endif

		// interrupt handling
		, input  RST_INT0
		, output INT0
		, input  RST_TIM0_OVF
		, output TIM0_OVF
		, input  RST_TIM0_COMPA	
		, output TIM0_COMPA	
		, input  RST_TIM0_COMPB
		, output TIM0_COMPB
		);
  
  parameter PORTA_ADDR =  0;   // PINA 0x00 , DDRA, PORTA
  parameter PORTB_ADDR =  4;   // PINB 0x04 , DDRB, PORTB
  parameter PORTC_ADDR = 27;   // PINC 0x1b , DDRC, PORTC
  parameter TIM0_ADDR  = 21;   // OCR0B 0x15 - TCCR0A 0x19
  parameter SREG_ADDR  = 63;   // SREG 0x3f , SPH,SPL,CCP,RSRFR,MCUCR 0x3a
  parameter MCUCR_ADDR = 58;   // MCUCR 0x3a
  parameter GIFR_ADDR  = 11;   // GIFR 0x0b , GIMSK 0x0c
  parameter TIFR_ADDR  = 37;   // TIFR 0x25 , TIMSK 0x26  
  parameter INT0_BIT = 18;      // PC2
  // parameter INT0_BIT = 0;      // PA0 : for test

`ifdef IOR_HAVE_SREG_SP
  reg  [7:0] i_sreg;
  reg  [7:0] i_sph;
  reg  [7:0] i_spl;

  assign FLAGS_OUT = i_sreg;
  assign SP_OUT  = { i_sph , i_spl };

  wire OE_sreg = RD & (ADDR == SREG_ADDR); 
  wire WE_sreg = WR & (ADDR == SREG_ADDR);
  wire OE_sph = RD & (ADDR == (SREG_ADDR -1)); 
  wire WE_sph = WR & (ADDR == (SREG_ADDR -1));
  wire OE_spl = RD & (ADDR == (SREG_ADDR -2)); 
  wire WE_spl = WR & (ADDR == (SREG_ADDR -2));
`else
  wire  [7:0] i_sreg = 0;
  wire  [7:0] i_sph = 0;
  wire  [7:0] i_spl = 0;

  wire OE_sreg = 0;
  wire OE_sph = 0;
  wire OE_spl = 0;
`endif

  wire OE_tifr = RD & (ADDR == TIFR_ADDR); 
  wire WE_tifr = WR & (ADDR == TIFR_ADDR);
  wire OE_timsk = RD & (ADDR == (TIFR_ADDR +1)); 
  wire WE_timsk = WR & (ADDR == (TIFR_ADDR +1));

  reg t0_ovf_prev;  
  reg t0_compa_prev;
  reg t0_compb_prev;  
  reg t0_ovf, t0_ovf_ie; // bit0
  reg t0_compa, t0_compa_ie; // bit1
  reg t0_compb, t0_compb_ie; // bit2

  assign TIM0_OVF   = t0_ovf & t0_ovf_ie;       // int 10
  assign TIM0_COMPA = t0_compa & t0_compa_ie ;  // int 11
  assign TIM0_COMPB = t0_compb & t0_compb_ie ;  // int 12

  //reg intf0, int0; // bit0

  wire OE_gifr = RD & (ADDR == GIFR_ADDR); 
  wire WE_gifr = WR & (ADDR == GIFR_ADDR);
  wire OE_gimsk = RD & (ADDR == (GIFR_ADDR +1)); 
  wire WE_gimsk = WR & (ADDR == (GIFR_ADDR +1));
  wire OE_mcucr = RD & (ADDR == MCUCR_ADDR); 
  wire WE_mcucr = WR & (ADDR == MCUCR_ADDR);

  wire [4:0] gi_int0_bit = INT0_BIT;
  reg gi_int0_in; 
  reg gi_int0_prev;  
  reg gi_int0, gi_int0_ie; // bit0
  reg [1:0] mcucr_isc;
  
  assign INT0 = gi_int0;

`ifdef OUTPUT_MUX_T1
    rtavr_ior_port #(.BASE_ADDR(PORTA_ADDR)) i_port_a(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(DO) , .PORT(PORTA)  
		, .WR(WR), .RD(RD), .ADDR(ADDR) );

    rtavr_ior_port #(.BASE_ADDR(PORTB_ADDR)) i_port_b(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(DO) , .PORT(PORTB)  
		, .WR(WR), .RD(RD), .ADDR(ADDR) );

    rtavr_ior_port #(.BASE_ADDR(PORTC_ADDR)) i_port_c(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(DO) , .PORT(PORTC) 
		, .WR(WR), .RD(RD), .ADDR(ADDR) );


    rtavr_ior_timer0 #(.BASE_ADDR(TIM0_ADDR)) i_timer0(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(DO)
		, .WR(WR), .RD(RD), .ADDR(ADDR)
		, .TIM0_OVF(TIM0_OVF_IN)
		, .TIM0_COMPB(TIM0_COMPB_IN)
		, .TIM0_COMPA(TIM0_COMPA_IN) );

  wire [7:0] i_DO = OE_sreg ? i_sreg : OE_sph ? i_sph : OE_spl ? i_spl 
                  : OE_tifr ? {5'b0 , t0_compb, t0_compa ,t0_ovf}
		  : OE_timsk ? {5'b0 , t0_compb_ie, t0_compa_ie ,t0_ovf_ie}
                  : OE_gifr ? {7'b0 , gi_int0}
		  : OE_gimsk ? {7'b0 , gi_int0_ie}
		  : OE_mcucr ? {mcucr_isc , 6'b0 }
		  : 0;

  bufif1(DO, i_DO, (OE_sreg | OE_sph | OE_spl | OE_tifr | OE_timsk ));

`else // not OUTPUT_MUX_T1

  wire [7:0] pa_DO;
  wire [7:0] pb_DO;
  wire [7:0] pc_DO;
  wire [7:0] t0_DO;
  
`ifdef IOR_HAVE_PORTA
    wire OE_pa = RD & ( (ADDR == PORTA_ADDR) | (ADDR == PORTA_ADDR+1)
		     | (ADDR == PORTA_ADDR+2) );
    rtavr_ior_port #(.BASE_ADDR(PORTA_ADDR)) i_port_a(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(pa_DO) , .PORT(PORTA)  
		, .WR(WR), .RD(RD), .ADDR(ADDR) );
`else
    wire OE_pa = 0;
`endif
`ifdef IOR_HAVE_PORTB
    wire OE_pb = RD & ( (ADDR == PORTB_ADDR) | (ADDR == PORTC_ADDR+1)
		     | (ADDR == PORTB_ADDR+2) );
    rtavr_ior_port #(.BASE_ADDR(PORTB_ADDR)) i_port_b(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(pb_DO) , .PORT(PORTB)  
		, .WR(WR), .RD(RD), .ADDR(ADDR) );
`else
    wire OE_pb = 0;
`endif

`ifdef IOR_HAVE_PORTC
    wire OE_pc = RD & ( (ADDR == PORTC_ADDR) | (ADDR == PORTC_ADDR+1)
		     | (ADDR == PORTC_ADDR+2) );
    rtavr_ior_port #(.BASE_ADDR(PORTC_ADDR)) i_port_c(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(pc_DO) , .PORT(PORTC) 
		, .WR(WR), .RD(RD), .ADDR(ADDR) );
`else
    wire OE_pc = 0;
`endif

   rtavr_ior_timer0 #(.BASE_ADDR(TIM0_ADDR)) i_timer0(
		.CLK(CLK), .RESET(RESET), .DI(DI), .DO(t0_DO)
		, .WR(WR), .RD(RD), .ADDR(ADDR)
		, .TIM0_OVF(TIM0_OVF_IN)
		, .TIM0_COMPB(TIM0_COMPB_IN)
		, .TIM0_COMPA(TIM0_COMPA_IN) );
   wire OE_t0 = RD & ( (ADDR == TIM0_ADDR) | (ADDR == TIM0_ADDR+1)
                     | (ADDR == TIM0_ADDR+2) | (ADDR == TIM0_ADDR+3)
		     | (ADDR == TIM0_ADDR+4) );

   wire [7:0] i_DO = OE_sreg ? i_sreg : OE_sph ? i_sph : OE_spl ? i_spl
                  : OE_tifr ? {5'b0 , t0_compb, t0_compa ,t0_ovf}
		  : OE_timsk ? {5'b0 , t0_compb_ie, t0_compa_ie ,t0_ovf_ie}
                  : OE_gifr ? {7'b0 , gi_int0}
		  : OE_gimsk ? {7'b0 , gi_int0_ie}
		  : OE_mcucr ? {mcucr_isc , 6'b0 }
		  : 0;
   assign DO =  (OE_sreg | OE_sph | OE_spl | OE_tifr | OE_timsk ) ? i_DO
              : OE_pa ? pa_DO : OE_pb ? pb_DO : OE_pc ? pc_DO
	      : OE_t0 ? t0_DO : 0;   

`endif // not OUTPUT_MUX_T1

  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
`ifdef IOR_HAVE_SREG_SP
	i_sreg <= 0;
	i_sph <= 0;
	i_spl <= 0;
`endif
	t0_compb_ie <= 0;
	t0_compa_ie <= 0;
	t0_ovf_ie <= 0;

	t0_ovf <= 0;
	t0_compa <= 0;

	t0_compb <= 0;

	t0_ovf_prev   <=  0;
	t0_compa_prev <=  0;
	t0_compb_prev <=  0;

	mcucr_isc <= 0;

	gi_int0_ie <= 0;
	gi_int0 <= 0;
	gi_int0_in   <= 0;
	gi_int0_prev   <=  0;
      end
    else
      begin
`ifdef IOR_HAVE_SREG_SP
	if(WE_sreg) i_sreg <= DI;
	else i_sreg <= FLAGS_IN;

	if(WE_sph) i_sph <= DI;
	else i_sph <= SP_IN[15:8];

	if(WE_spl) i_spl <= DI;
	else i_spl <= SP_IN[7:0];
`endif
	if(WE_timsk) {t0_compb_ie, t0_compa_ie ,t0_ovf_ie} <= DI[2:0];

	if(RST_TIM0_OVF) t0_ovf <= 0;
	else if((WE_tifr & DI[0]) 
             | (t0_ovf_prev ^ TIM0_OVF_IN) & ~t0_ovf_prev ) t0_ovf <= 1;

	if(RST_TIM0_COMPA) t0_compa <= 0;
	else if((WE_tifr & DI[1])
             | (t0_compa_prev ^ TIM0_COMPA_IN) & ~t0_compa_prev) t0_compa <= 1;

	if(RST_TIM0_COMPB) t0_compb <= 0;
	else if((WE_tifr & DI[2])
             | (t0_compb_prev ^ TIM0_COMPB_IN) & ~t0_compb_prev) t0_compb <= 1;

	t0_ovf_prev   <=  TIM0_OVF_IN;
	t0_compa_prev <=  TIM0_COMPA_IN;
	t0_compb_prev <=  TIM0_COMPB_IN; 

	if(WE_mcucr) {mcucr_isc} <= DI[7:6];

	if(WE_gimsk) {gi_int0_ie} <= DI[0:0];
	if(RST_INT0) gi_int0 <= 0;
	else if((WE_gifr & DI[0])
              | (gi_int0_prev ^ gi_int0_in) 
                 & ((mcucr_isc == 1) 
                   | (mcucr_isc == 2) & ( gi_int0_prev) 
                   | (mcucr_isc == 3) & ~( gi_int0_prev) )  
	       | ((mcucr_isc == 0) & ~gi_int0_in) ) gi_int0 <= 1;

	gi_int0_in   <= 
`ifdef IOR_HAVE_PORTA
	( gi_int0_bit[4:3] == 0) ? PORTA[gi_int0_bit[2:0]] :
`endif
`ifdef IOR_HAVE_PORTB
	( gi_int0_bit[4:3] == 1) ? PORTB[gi_int0_bit[2:0]] :
`endif
`ifdef IOR_HAVE_PORTC
	( gi_int0_bit[4:3] == 2) ? PORTC[gi_int0_bit[2:0]] :
`endif
	0;

	gi_int0_prev   <=  gi_int0_in;
      end
  end

endmodule
