/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"
// global defines : OUTPUT_MUX_T1
//`define OUTPUT_MUX_T1

// Module Name: rtavr_rom
//  ver 003
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                   (1KW)       NORMAL    TRISTATE(FAKE)   
// Number of Slice Flip Flops:      8          
// Number of 4 input LUTs:         12         
// Number of occupied Slices:      11          
// Total Number of 4 input LUTs:   12         
// Number of bonded IOBs :         57         57
//   IOB Flip Flops :              -          -
// Number of BUFGMUXs :             1          1
// Number of RAMB16BWEs :           1          1
// Clock to Setup 
//on destination clock CLK      1.905       1.900
// WARNING:Xst:647 - Input <ADDRA<11:10>> is never used. ..
// WARNING:Xst:646 - Signal <i_addrb<11:10>> is assigned but ..

module rtavr_rom # (
		parameter SIZE = 1024
//		parameter SIZE = 2048
//		parameter SIZE = 4096
) (
		input CLK,
		input RESET,

		input [11:0] ADDRA,
		output [15:0] DOA,

		input WEB,OE, 
		input [12:0] ADDRB,
		input [7:0] DIB, 
		output[7:0] DOB
		);
		
  reg  [15:0] mem [0:SIZE-1];
  reg  [15:0] i_doa;
  assign DOA = i_doa;

  always @(posedge CLK)
  begin
	i_doa <= mem[ADDRA];
  end

  reg  [7:0] i_tmp;
  reg  [15:0] i_dob;
  
  wire [11:0] i_addrb = ADDRB[12:1]; 
  wire WEBH = ADDRB[1:0] & WEB;
  wire WEBL = ~ADDRB[1:0] & WEB;

`ifdef OUTPUT_MUX_T1
  bufif1(DOB, (ADDRB[1:0])? i_dob[15:8]: i_dob[7:0], OE);
`else 
  assign DOB = ~OE ? 0 : (ADDRB[1:0])? i_dob[15:8]: i_dob[7:0];
`endif  
  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	i_tmp <=  0;
	i_dob <= 0;
      end
    else
      begin
	if(WEBL) i_tmp <=  DIB;
	if(WEBH) mem[i_addrb] <=  { DIB , i_tmp };
	i_dob <= mem[i_addrb];
      end
  end

endmodule
