/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"

//`define S0_INCLUDE_ROM
// Module Name: rtavr_s0_fetch
//   Ver 002
//
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//            -  with 8KB ROM -   NORMAL(new) NORMAL(old)  NO_SREG  NO_EXACC
// Number of Slice Flip Flops:       24         24           24     24
// Number of 4 input LUTs:           50         44           44     39
// Number of occupied Slices:        39         36           35     33
// Total Number of 4 input LUTs:     51         45           44     39 
// Number of bonded IOBs :           91         91           91     84
//   IOB Flip Flops :                 -          -            -      -
// Number of BUFGMUXs :               1          1            1      1
// Number of RAMB16BWEs :             4          4            4      4
// Clock to Setup 
//on destination clock CLK        4.382      4.101         3.832  3.660
// NORMAL(new) : add i_rr (use PORTB for all store inst. )

module rtavr_s0_fetch (
		input CLK,
		input RESET,
`ifdef S0_INCLUDE_ROM
		input [11:0] PC,
		input WEB, OE, 
		input [12:0] ADDRB,
		input [7:0] DIB, 
		output[7:0] DOB,
`else
		input [15:0] PM_OUT,
`endif
		input WB_EN,
		output WB_INST,
		output [15:0] INST,
		output[3:0] GPR_ADDRAL,
		output[3:0] GPR_ADDRAH,
		output[3:0] GPR_ADDRBL,
		output[3:0] GPR_ADDRBH,
		output GPR_PREDEC, GPR_POSTINC
		);

  reg [15:0] i_inst;
  assign INST = i_inst;
`ifdef S0_INCLUDE_ROM
  wire [15:0] PM_OUT;
  rtavr_rom_8KB PM(.CLK(CLK) , .ADDRA(PC), .DOA(PM_OUT) , 
		 .WEB(WEB) , .OE(OE) , .ADDRB(ADDRB),
		 .DIB(DIB), .DOB(DOB) 
		);
`endif

  wire i_predec = ( PM_OUT[15:10] == 6'b100100 ) & ( PM_OUT[1:0] == 2'b10 );
//         LD.Rd.--X	1001:000d:dddd:1110
//         ST.--X.Rr	1001:001r:rrrr:1110
//         LD.Rd.--Y	1001:000d:dddd:1010
//         ST.--Y.Rr	1001:001r:rrrr:1010
//         LD.Rd.--Z	1001:000d:dddd:0010
//         ST.--Z.Rr	1001:001r:rrrr:0010
//         ELPM.Rd.Z    1001:000d:dddd:0110  << undefined instruction
  wire i_postinc = ( PM_OUT[15:10] == 6'b100100 ) & ( PM_OUT[1:0] == 2'b01 );
//         LD.Rd.X++	1001:000d:dddd:1101
//         ST.X++.Rr	1001:001r:rrrr:1101
//         LD.Rd.Y++	1001:000d:dddd:1001
//         ST.Y++.Rr	1001:001r:rrrr:1001
//         LD.Rd.Z++	1001:000d:dddd:0001
//         ST.Z++.Rr	1001:001r:rrrr:0001
//         LPM.Rd.Z++   1001:000d:dddd:0101  << undefined instruction
  wire i_xreg = ( PM_OUT[15:10] == 6'b100100 ) & ( PM_OUT[3:2] == 2'b11 )
                  & ( PM_OUT[1:0] != 2'b11 );
//       0xa,0xb: (26 - 16, 27 - 16)
//         LD.Rd.X	1001:000d:dddd:1100
//         ST.X.Rr	1001:001r:rrrr:1100
//         LD.Rd.X++	1001:000d:dddd:1101
//         ST.X++.Rr	1001:001r:rrrr:1101
//         LD.Rd.--X	1001:000d:dddd:1110
//         ST.--X.Rr	1001:001r:rrrr:1110
//         POP          1001:000d:dddd:1111  <<<
//         PUSH         1001:001d:dddd:1111  <<<
  wire i_yreg = ( PM_OUT[15:10] == 6'b100100 ) & ( PM_OUT[3:2] == 2'b10 );
//       0xc,0xd:(28 - 16, 29 - 16)
//         LD.Rd.Y	1000:000d:dddd:1000
//         ST.Y.Rr	1000:001r:rrrr:1000
//         LD.Rd.Y++	1001:000d:dddd:1001
//         ST.Y++.Rr	1001:001r:rrrr:1001
//         LD.Rd.--Y	1001:000d:dddd:1010
//         ST.--Y.Rr	1001:001r:rrrr:1010
  wire i_zreg = ( PM_OUT[15:10] == 6'b100100 ) & ( PM_OUT[3:2] == 2'b00 );
//       0xe,0xf:(30 - 16, 31 - 16)
//         LD.Rd.Z	1000:000d:dddd:0000
//         ST.Z.Rr	1000:001r:rrrr:0000
//         LD.Rd.Z++	1001:000d:dddd:0001
//         ST.Z++.Rr	1001:001r:rrrr:0001
//         LD.Rd.--Z	1001:000d:dddd:0010
//         ST.--Z.Rr	1001:001r:rrrr:0010
  
  wire i_icall = ( PM_OUT[15:10] == 6'b100101 ) & PM_OUT[3] & PM_OUT[0];
//        0xe,0xf:(30 - 16, 31 - 16)
//          IJMP	1001:0100:0000:1001
//          ICALL	1001:0101:0000:1001
//          EIJMP       1001:0100:0001:1001  << undefined instruction
//          EICALL      1001:0101:0001:1001  << undefined instruction

  wire i_rr = ( PM_OUT[15] == 1 );
//         STS		1001:001d:dddd:0000 
//         PUSH		1001:001d:dddd:1111 
//         SBRC		1111:110r:rrrr:0bbb 
//         SBRS		1111:111r:rrrr:0bbb
//         OUT          1011:1AAr:rrrr:AAAA 
//           :
//         MUL          1001:11rd:dddd:rrrr  << undefined instruction
  
assign WB_INST  =  i_predec | i_postinc;
assign GPR_PREDEC  = WB_EN & i_predec;
assign GPR_POSTINC = WB_EN & i_postinc;
assign GPR_ADDRAL   = i_xreg ? (26 - 16) : i_yreg ? (28 - 16)
                    : i_zreg ? (30 - 16) : PM_OUT[7:4];
assign GPR_ADDRAH   = i_xreg ? (27 - 16) : i_yreg ? (29 - 16)
                    : i_zreg ? (31 - 16) : INST[7:4];
assign GPR_ADDRBL   = i_icall ? (30 - 16) : PM_OUT[7:4];
assign GPR_ADDRBH   = i_icall ? (31 - 16) : i_rr ? PM_OUT[7:4] : PM_OUT[3:0];

  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	i_inst <= 0;
      end
    else
      begin
        i_inst <= PM_OUT;
      end
  end

endmodule