/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"

// Module Name: rtavr_s1_decode
//   Ver 002
//
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                                NORMAL
// Number of Slice Flip Flops:       17
// Number of 4 input LUTs:           10     
// Number of occupied Slices:        10    
// Total Number of 4 input LUTs:     17   
// Number of bonded IOBs :           47  
//   IOB Flip Flops :                33 
// Number of BUFGMUXs :               1
// Number of RAMB16BWEs :             - 
// Clock to Setup 
//on destination clock CLK            ?

module rtavr_s1_decode (
		input CLK,
		input RESET,
		input [15:0] INST,
		input [7:0] FLAGS_IN,
		input [11:0] JA,
		input WB_VALID,

		input S0_WB_INST,
		output S0_WB_EN,

		output CMD_LOAD, CMD_STORE,
		output CMD_LOAD_WR,
		output CMD_OP3, CMD_OP2, CMD_OP1,
		output [2:0] OP3_SEL,
		output OP3_WC, CMD_OP3_WR, 
		output [2:0] OP2_SEL,
		output [2:0] OP1_SEL,
		output [7:0] IMM_DATA,

		output CMD_SBIX,
		output CMD_SBRX,
		output [2:0] SBIX_BIT,
		input SBIX_BIT_IN,
		
		output EA_ABS,
		output EA_PUSH, EA_POP,
		output [1:0] PUSH_SEL,
		output [7:0] ABS_ADDR,
		output [11:0] PC
		);
  reg n_invalid;

  reg i_cmd_load;
  reg i_cmd_load_wr;
  reg i_cmd_store;
  reg i_cmd_op3;
  reg i_cmd_op3_wr;
  reg i_cmd_op2;
  reg i_cmd_op1;
  reg [2:0] i_op3_sel;
  reg i_op3_wc;
  reg [2:0] i_op2_sel;
  reg [2:0] i_op1_sel;

  reg i_ea_abs; // IN,OUT, LDS,STS
  reg [7:0] i_ea;
 
  reg i_ea_push;
  reg i_ea_pop;
  reg [1:0] i_push_sel;

  reg s2_cpse;
  reg s2_sbix;
  reg s2_sbix_sel;
  reg [2:0] s2_sbix_bit;
  reg s2_sbrx;

  reg [7:0] i_imm;
  reg [11:0] i_pc;
 
  assign S0_WB_EN = 1'b1;
 
  assign CMD_LOAD = i_cmd_load;
  assign CMD_LOAD_WR = i_cmd_load_wr;
  assign CMD_STORE = i_cmd_store;

  assign CMD_OP3 = i_cmd_op3;
  assign CMD_OP3_WR = i_cmd_op3_wr;
  assign OP3_SEL = i_op3_sel;
  assign OP3_WC = i_op3_wc;
  
  assign CMD_OP2 = i_cmd_op2;
  assign OP2_SEL = i_op2_sel;
  assign CMD_OP1 = i_cmd_op1;
  assign OP1_SEL = i_op1_sel;

  assign EA_ABS = i_ea_abs;
  assign EA_PUSH = i_ea_push;
  assign EA_POP = i_ea_pop;
  assign PUSH_SEL = i_push_sel;
  assign ABS_ADDR = i_ea;

  assign CMD_SBIX = s2_sbix;
  assign CMD_SBRX = s2_sbrx;
  assign SBIX_BIT = s2_sbix_bit;
  
  assign IMM_DATA = i_imm;
  assign PC = i_pc;

// EA_ABS / ABS_ADDR
// IN              1011:0AA1:dddd:AAAA     # 00
// OUT             1011:1AA1:rrrr:AAAA     # 00
// LDS             1010:0kkK:dddd:kkkk     # 
// STS             1010:1kkK:dddd:kkkk     # 
//                 101X:S--X----------     # 
//  INST[12]/INST[8]  -> v_ea_abs[7:6]
//                  01        10
//                  00        01
//                  10        11 (undefined)
//                  11        00
// CBI             1001:1000:AAAA:Abbb     # 
// SBI             1001:1010:AAAA:Abbb     #
// SBIC            1001:1001:AAAA:Abbb     #
// SBIS            1001:1011:AAAA:Abbb     # 

  wire v_ea_abs = (INST[15:13] == 3'b101)
		  | (INST[15:10] == 6'b100110) ;

  wire [7:0] v_ea = (INST[15:13] == 3'b101)
		? { INST[12] ^ INST[8] ,  ~INST[8] , INST[10:9] , INST[3:0] }
		: { 3'b000 , INST[7:3] } ;

 wire f_sbix = (INST[15:10] == 6'b100110) & INST[8];
 
// # OP3 INST
// NOP		   0000:0000:0000:0000	   #
// CPSE            000B:00rd:dddd:rrrr     # not LOAD 
// CP/CPC          000B:01rd:dddd:rrrr     #
// SUB/SBC         000B:10rd:dddd:rrrr     #
// ADD/ADC         000C:11rd:dddd:rrrr     # 
// AND             0010:00rd:dddd:rrrr     # 
// EOR             0010:01rd:dddd:rrrr     # 
// OR              0010:10rd:dddd:rrrr     # 
// MOV             0010:11rd:dddd:rrrr     # 
//                 --*-:**------------
// CPI             0011:KKKK:dddd:KKKK     # 
// MOVW            0000:0001:dddd:rrrr     # undefined inst
// MULS            0000:0010:dddd:rrrr     # undefined inst
// MULSU           0000:0011:0ddd:0rrr     # undefined inst
// FMUL            0000:0011:0ddd:1rrr     # undefined inst
// FMULS           0000:0011:1ddd:0rrr     # undefined inst
// FMULSU          0000:0011:1ddd:1rrr     # undefined inst

  wire v_cmd_op3 = (INST[15:14] == 2'b00) & ~(INST[13:12] == 2'b11);
  wire v_cmd_op3_wr = v_cmd_op3 & (INST[13] | ~(INST[11:10] == 2'b00));
  wire f_cpse = v_cmd_op3 & ~INST[13] & (INST[11:10] == 2'b00);
  wire [2:0] v_op3_sel = { INST[13] , INST[11:10] };
  wire v_op3_wc = INST[12];

  wire f_skip = ( s2_cpse | s2_sbrx | s2_sbix ) & s2_sbix_sel & SBIX_BIT_IN
               | ( s2_sbrx | s2_sbix ) & ~s2_sbix_sel & ~SBIX_BIT_IN ;
 
// # OP2 INST
// LDI             1110:KKKK:dddd:KKKK     # OP2_SEL = 2
// CPI             0011:KKKK:dddd:KKKK     # OP2_SEL = 3
// SBCI            0100:KKKK:dddd:KKKK     # 
// SUBI            0101:KKKK:dddd:KKKK     # 
// ORI             0110:KKKK:dddd:KKKK     # 
// ANDI            0111:KKKK:dddd:KKKK     # 
//                 -***:--------------      

  wire v_cmd_op2 = (INST[15:14] == 2'b01 )
			| (INST[15:12] == 4'b0011)
			| (INST[15:12] == 4'b1110) ;

  wire [2:0] v_op2_sel = { (INST[15] ^ INST[14]), INST[13:12] };
  wire [7:0] v_imm = { INST[11:8] , INST[3:0] };

// # OP1 INST
// COM             1001:010d:dddd:0000     # 
// NEG             1001:010d:dddd:0001     # 
// SWAP            1001:010d:dddd:0010     # 
// INC             1001:010d:dddd:0011     # 
// DEC             1001:010d:dddd:1010     # OP1_SEL = 4
// ASR             1001:010d:dddd:0101     # 
// LSR             1001:010d:dddd:0110     # 
// ROR             1001:010d:dddd:0111     # 
//                 ----:-----------***      

  wire f_ext = (INST[15:9] == 7'b1001010 );
  wire v_cmd_op1 = f_ext & (~INST[3] | (INST[3:0] == 4'b1010));

  wire [2:0] v_op1_sel = (INST[3]) ? 3'b100 : INST[2:0];

// LOAD Rr (Rd -> Rr @s0)
// LD.Rd.Z         1000:000d:dddd:0000     # 
// LD.Rd.Y         1000:000d:dddd:1000     # 
// LD.Rd.Z++       1001:000d:dddd:0001     # 
// LD.Rd.--Z       1001:000d:dddd:0010     # 
// LD.Rd.Y++       1001:000d:dddd:1001     # 
// LD.Rd.--Y       1001:000d:dddd:1010     # 
// LD.Rd.X         1001:000d:dddd:1100     # 
// LD.Rd.X++       1001:000d:dddd:1101     # 
// LD.Rd.--X       1001:000d:dddd:1110     # 

// POP             1001:000d:dddd:1111     # 

// SYORE Rr
// ST.Z.Rr         1000:001r:rrrr:0000     # 
// ST.Y.Rr         1000:001r:rrrr:1000     # 
// ST.Z++.Rr       1001:001r:rrrr:0001     # 
// ST.--Z.Rr       1001:001r:rrrr:0010     # 
// ST.Y++.Rr       1001:001r:rrrr:1001     # 
// ST.--Y.Rr       1001:001r:rrrr:1010     # 
// ST.X.Rr         1001:001r:rrrr:1100     # 
// ST.X++.Rr       1001:001r:rrrr:1101     # 
// ST.--X.Rr       1001:001r:rrrr:1110     # 

// PUSH            1001:001d:dddd:1111     # 
//                 100X:00S-----------

// LPM.Rd.Z        1001:000d:dddd:0100     # undefined inst
// LPM.Rd.Z++      1001:000d:dddd:0101     # undefined inst
// ELPM.Rd.Z       1001:000d:dddd:0110     # undefined inst
// ELPM.Rd.Z++     1001:000d:dddd:0111     # undefined inst
// LDS(2word)      1001:000d:dddd:0000     # undefined inst
// STS(2word)      1001:001d:dddd:0000     # undefined inst

  // LD/ST + POP/PUSH
  wire v_ldst = (INST[15:13] == 3'b100) & (INST[11:10] == 2'b00);
  
// CBI             1001:1000:AAAA:Abbb     # 
// SBI             1001:1010:AAAA:Abbb     #
// SBIC            1001:1001:AAAA:Abbb     #
// SBIS            1001:1011:AAAA:Abbb     # 

  wire v_ldio = (INST[15:10] == 6'b100110);
  wire v_stio = v_ldio & ~INST[8];

//
// IN              1011:0AA1:dddd:AAAA     # 
// LDS             1010:0kkK:dddd:kkkk     # 
// OUT             1011:1AA1:rrrr:AAAA     # 
// STS             1010:1kkK:dddd:kkkk     #
//                 101-:S-------------

  wire v_cmd_load_wr = ( v_ldst & ~INST[10] )  // LD,POP
                    | ( v_ea_abs & ~INST[11] ) // IN, LDS
		    |  v_stio ;                // CBI,SBI

  wire v_cmd_load = ( v_ldst & ~INST[10] )     // LD,POP
                    | ( v_ea_abs & ~INST[11] ) // IN, LDS
  		    |  v_ldio ;                // CBI,SBI , SBIC,SBIS
		    
  wire v_cmd_store = v_ldst & INST[10]           // ST,PUSH
                    | ( v_ea_abs & INST[11] );   // OUT, STS


// IJMP            1001:0100:0000:1001     # 
// ICALL           1001:0101:0000:1001     # 
//                 EEEE:EEE*:----:1--1     # 

  wire f_ij =  f_ext & (~INST[8] & INST[3] & INST[0]);
  wire f_ic =  f_ext & (INST[8] & INST[3] & INST[0]);

// RJMP            1100:kkkk:kkkk:kkkk     # 
// RCALL           1101:kkkk:kkkk:kkkk     # 

  wire [11:0] adr_rj = INST[11:0];
  wire f_rj =  (INST[15:12] == 4'b1100 );
  wire f_rc =  (INST[15:12] == 4'b1101 );

// PUSH            1001:001d:dddd:1111     # 
// ICALL           1001:0101:0000:1001     # 
// RCALL           1101:kkkk:kkkk:kkkk     # 

// POP             1001:000d:dddd:1111     # 
// RET             1001:0101:0000:1000    # 
// RETI            1001:0101:0001:1000    # 
//                 EEEE:EEE1:0--*:1--0    #

  wire f_ret = f_ext & (INST[8:7] == 2'b10) & INST[3] & ~INST[0]; // RET/RETI
  wire f_ret_sel  = INST[4];                                      // is RETI

  wire v_push_pop = v_ldst & (INST[3:0] == 4'b1111);
  wire v_ea_pop = ( v_push_pop & ~INST[10] )   // POP;
                   | f_ret ;                   // RET,RETI
  wire v_ea_push = ( v_push_pop & INST[10] )   // PUSH;
                   | f_rc | f_ic;              // RCALL,ICALL

  wire v_push_sel = ( f_ret | f_rc | f_ic )? 2'b10    // PCL
                    : 2'b00;                          // GPR

// BLD             1111:100d:dddd:0bbb     #
// BST             1111:101d:dddd:0bbb     # 

// BRBC            1111:01kk:kkkk:ksss     # 
// BRBS            1111:00kk:kkkk:ksss     #

  wire [11:0] adr_br = { INST[9], INST[9], INST[9], INST[9], INST[9]
			 ,INST[9:3] };
  wire f_br = (INST[15:11] == 5'b11110 );

// BSET            1001:0100:0sss:1000     # 
// BCLR            1001:0100:1sss:1000     # 
  
// CBI             1001:1000:AAAA:Abbb     # 
// SBI             1001:1010:AAAA:Abbb     #

// SBIC            1001:1001:AAAA:Abbb     #
// SBIS            1001:1011:AAAA:Abbb     #

// SBRC            1111:110r:rrrr:0bbb     #
// SBRS            1111:111r:rrrr:0bbb     #

  wire f_sbrx = (INST[15:10] == 6'b111111 );
  wire f_sbix_sel = INST[9]; // 0 : SBIC/SBRC 1: SBIS/SBRS
  wire [2:0] f_sbix_bit = INST[2:0];
  
  wire s_stall = WB_VALID ;
  wire s_invalid = n_invalid;
  wire s0_stall = s_stall | f_ij | f_ic | f_br | f_rj | f_rc;
  wire [11:0] v_pc = f_ij ? JA : i_pc + 1 
                         + ( f_rj ? adr_rj 
                           : f_br ? adr_br 
		           : 0);
 
// SLEEP           1001:0101:1000:1000     #
// WDR             1001:0101:1010:1000     #
// BREAK           1001:0101:1001:1000     # undefined inst
//                 EEEE:EEE1:1-**:1--0     #

  wire f_mcu = f_ext & (INST[8:7] == 2'b11) & INST[3] & ~INST[0];
  wire f_sleep = f_mcu & (INST[5:4] == 2'b00);
  wire f_wdr   = f_mcu & (INST[5:4] == 2'b10);

  always @(posedge CLK)
  begin
    if (RESET == 0) 
      begin
	n_invalid     <= 0;

	i_pc          <= 0;
	i_imm         <= 0;

	i_ea          <= 0;
	i_ea_abs      <= 0;
	i_ea_push     <= 0;
	i_ea_pop      <= 0;
	i_push_sel    <= 0;

	s2_cpse        <= 0;
	s2_sbix        <= 0;
	s2_sbrx        <= 0;
	s2_sbix_sel    <= 0;
	s2_sbix_bit    <= 0;

	i_cmd_op3     <= 0;
	i_cmd_op3_wr  <= 0;
	i_op3_sel     <= 0;
	i_op3_wc      <= 0;

	i_cmd_op2     <= 0;
	i_op2_sel     <= 0;

	i_cmd_op1     <= 0;
	i_op1_sel     <= 0;

	i_cmd_load    <= 0;
	i_cmd_load_wr <= 0;
	i_cmd_store   <= 0;
      end
    else
      begin
	n_invalid     <= s_stall | f_ij | f_ic | f_br | f_rj | f_rc | f_skip;
	if (~s_stall) i_pc <= v_pc;
	i_imm         <= v_imm;

	i_ea          <= v_ea;
	i_ea_abs      <= v_ea_abs;
	i_ea_push     <= v_ea_push;
	i_ea_pop      <= v_ea_pop;
	i_push_sel    <= v_push_sel;

	s2_cpse        <= f_cpse;
	s2_sbix        <= f_sbix;
	s2_sbrx        <= f_sbrx;
	s2_sbix_sel    <= f_sbix_sel;
	s2_sbix_bit    <= f_sbix_bit;

	i_cmd_op3     <= ~s_invalid & v_cmd_op3;
	i_cmd_op3_wr  <= ~s_invalid & v_cmd_op3_wr;
	i_op3_sel     <= v_op3_sel;
	i_op3_wc      <= v_op3_wc;

	i_cmd_op2     <= ~s_invalid & v_cmd_op2;
	i_op2_sel     <= v_op2_sel;

	i_cmd_op1     <= ~s_invalid & v_cmd_op1;
	i_op1_sel     <= v_op1_sel;

	i_cmd_load    <= ~s_invalid & v_cmd_load;
	i_cmd_load_wr <= ~s_invalid & v_cmd_load_wr;
	i_cmd_store   <= ~s_invalid & v_cmd_store;
      end
  end

endmodule