/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"

// global defines : OUTPUT_MUX_T1

//`define OUTPUT_MUX_T1

// Module Name: rtavr_ram_2KB
//  ver 003
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                          DO:  NORMAL    TRISTATE(FAKE)   
// Number of Slice Flip Flops:      9          0   
// Number of 4 input LUTs:          5          0
// Number of occupied Slices:       5          0
// Total Number of 4 input LUTs:    8          0 
// Number of bonded IOBs :         59         59
//   IOB Flip Flops :              -          -
// Number of BUFGMUXs :             1          1
// Number of RAMB16BWEs :           1          1
// Clock to Setup 
//on destination clock CLK          ?          ?

module rtavr_sram_2KB(
		input CLK,
		input RESET,

		input WEA,
		input [10:0] ADDRA,
		input [7:0] DIA,
		output [7:0] DOA,

		input WEB, OE,
		input [10:0] ADDRB,
		input [7:0] DIB,
		output [7:0] DOB
		);
		
  reg  [7:0] mem [0:2048-1];
  reg  [7:0] i_doa;
  assign DOA = i_doa;

  always @(posedge CLK)
  begin
    if(WEA) mem[ADDRA] <= DIA;
    i_doa <= mem[ADDRA];
  end

  reg  [7:0] i_dob;

`ifdef OUTPUT_MUX_T1
  bufif1(DOB, i_dob, OE);
`else
  assign DOB = OE? i_dob : 0;
`endif
  always @(posedge CLK)
  begin
    if (RESET == 0)
      begin
	i_dob <= 0;
      end
    else
      begin
	if(WEB) mem[ADDRB] <= DIB;
	i_dob <= mem[ADDRB];
      end
  end

endmodule
